-- mirror_outlet_state.lua
--  A pdu observes the outlet state from outlet 1 to 11 and the outlet named 'main a'.
--  If an outlet has changed the state the script mirrored the state to a remote pdu
--  (1 to 11, main a)

-- NOTE: with some extra comments
-- NOTE: start reading from 'start reading here'!
-- NOTE: lua arrays are beginning from position one (one-based-indexing)
--       and idl arrays are starting from position zero (zero-based-indexing)
-- NOTE: a remote call costs time, you should limit it
-- NOTE: instead of assert() should you implement in a production version a
--       better exception handling

require "Pdu"
require "EventService"

-- to switch on/off printing
local function printInfo(...)
  if true then     -- set to true for printing and to false for no printing
	  print("INFO:", (...))
  end
end

-- function to find the first outlet at pdu referred by name
-- returns the position in the getOutlets-array
local function searchOutletByName(pdu, name)
  local outlets = pdu:getOutlets()  -- outlet array
  for position=1,#outlets do        -- from first to the last outlet
    printInfo("get settings for outlet " .. position)
  	local settings = outlets[position]:getSettings()
  	if settings.name == name then   -- is true when the name matches
  		printInfo("outlet is at position " .. position)
  		return position  -- return the position
  	end
  end
  printInfo("can't find output " .. name)
  return nil           -- on failure return nil
end

-- function to iterate through the global devices list to find the sensor
-- and the remote outlet position
local function getSensorAndPosition(sensor)
  for i=1,#devices do  -- go through global devices table and find the sensor
    local stateSensor = sensors.StateSensor:cast(sensor.source)  -- cast sensor.source to stateSensor
    if stateSensor:equals(devices[i]) then  -- compare the remote object and local if they equals
      local rPos = remoteOutletPositions[i]
      return stateSensor, rPos
    end
  end
  return nil, nil  -- if there is no match
end

-- function that processes a state sensor change event
local function processStateSensorChanging(sensor)
  printInfo("state changed " .. sensor.newState.value .. " from sensor " .. sensor.source:getId())
  local stateSensor, rPos = getSensorAndPosition(sensor)
  if stateSensor then
  	printInfo("will switch remote outlet " .. rPos)
    -- a remote outlet object: well known url, agent.HttpAgent
    -- lua-array != idl-array  =>  decrement rPos by one
  	local outlet = pdumodel.Outlet:newRemote("/model/pdu/0/outlet/" .. rPos-1, httpAgent)
  	local ret = outlet:setPowerState( sensor.newState.value )
    -- if no failure happens, the return value is zero
  	assert(ret == 0)
  else
    assert(0)  -- should not happen
  end
end

-- start reading here
printInfo("mirror outlet state is starting")

printInfo("create a remote pdu")
 -- a new http agent object with: ip addr, user, password
 -- NOTE: agent = agent.HttpAgent:new() overrides the agent table!
httpAgent = agent.HttpAgent:new("192.168.7.15", "admin", "raritan")
 -- a new remote pdu object with: well known uri, http agent
local remotePdu = pdumodel.Pdu:newRemote("/model/pdu/0", httpAgent)

-- outlet name
local name = "main a"
printInfo("try to find outlet with name \'" .. name .. "\' on the remote pdu")
local remotePosition = searchOutletByName(remotePdu, name)
assert(remotePosition)  -- abort when there is no output

printInfo("create a local pdu")
-- a new local pdu object
local localPdu  = pdumodel.Pdu:getDefault()
printInfo("find outlet with name \'" .. name .. "\' on the local pdu")
local localPosition = searchOutletByName(localPdu, name)
assert(localPosition)

-- create a table with all devices to observe and mirror the local state form the
-- desired outputs to the remote outputs
local outlets = localPdu:getOutlets()
-- a global table with all local outlet positions
localOutletPositions = {}
for i=1,11 do localOutletPositions[i] = i end
localOutletPositions[12] = localPosition

devices = {} -- createS a global empty table to fill with the first elevens outlets
for i=1,11 do
	local sensors = outlets[i]:getSensors() -- get all sensors from outlet i
  devices[i] = sensors.outletState        -- copy only the outlet state sensor
end
devices[12] = outlets[localPosition]:getSensors().outletState  -- same as above in one line

-- a global table with all remote outlet positions
remoteOutletPositions = {}
for i=1,11 do remoteOutletPositions[i] = i end
remoteOutletPositions[12] = remotePosition

printInfo("mirroring actually states")
for i=1,12 do
	local rPos = remoteOutletPositions[i]
	local lPos = localOutletPositions[i]
  -- remote outlet position (zero-based-indexing!!!)
	local outlet = pdumodel.Outlet:newRemote("/model/pdu/0/outlet/" .. rPos-1, httpAgent)
	local pwrState = outlets[lPos]:getState().powerState
  local ret = outlet:setPowerState(pwrState)
	assert(ret == 0)  -- abort on failure
  -- lPos and rPos can be used as 'label'
  printInfo("mirrow local outlet " .. lPos .. " to remote outlet " .. rPos )
end

local outlets = nil -- outlets is no more needed => throw away

-- create a new event service object and a new channel
local service = event.Service:getDefault()
local channel = service:createChannel()

-- register all devices (sensors) for a special event
for i=1,#devices do
  channel:demandEvent(sensors.StateSensor.StateChangedEvent.TYPE_ID, devices[i])
end

printInfo("start observing ...")
-- for ever loop
while true do
  local more, events = channel:pollEvents()  -- the variable more will be ignored
  if #events > 0 then
  	printInfo("received " .. #events .. " events")
  	for i=1,#events do
  		-- abort if tpye id is wrong
  	  assert(events[i].TYPE_ID == sensors.StateSensor.StateChangedEvent.TYPE_ID)
  	  processStateSensorChanging(events[i])  	
    end
  else
  	printInfo("received no events")
  end
end
